import java.math.BigInteger;
import java.util.ArrayList;
import java.util.Calendar;

/**
 * Eine Primzahlenverwaltung.
 * 
 * @author Q11/Q12 
 * @version 2015-03-25
 */
public class Primzahlenverwaltung
{
    /*
     * Die Liste der uns bisher bekannten Primzahlen.
     */
	private ArrayList<BigInteger> p;
	
	/*
	 * Die groesste uns bisher bekannte Primzahl.
	 */
	private BigInteger groesste;
	
	/*
	 * Anfangs- und Endzeit von Vorgaengen.
	 */
	private Calendar anfangszeit;
	private Calendar endzeit;

	/**
	 * Konstruktor.
	 */
	public Primzahlenverwaltung ()
	{
	    /*
	     * Beginne mit einer Primzahlliste, die nur aus der Zahl 2 besteht.
	     */
		p = new ArrayList<BigInteger> ();
		groesste = new BigInteger ("2");
		p.add (groesste);
	}
 
    /**
     * Ist die angegebene Zahl eine Primzahl?
     */
    public boolean istPrimzahl (String zahlAlsText)
    {
        schalteStoppuhrAn ();
        
        BigInteger zahl = new BigInteger (zahlAlsText);
        
        boolean antwort = istPrimzahl (zahl);
        
        if (antwort)
        {
            System.out.println (zahl + " ist prim.");
        }
        else
        {
            System.out.println (zahl + " ist nicht prim.");
        }
        
        schalteStoppuhrAus ();
        
        return antwort;
    }
    
    private boolean istPrimzahl (BigInteger zahl) 
    {
        if (gibPrimfaktor (zahl).equals (zahl))
        {
            return true;
        }
        else
        {
            return false;
        }
    }
 
    /**
     * Zerlege die angegebene Zahl in Primfaktoren.
     */
    public void faktorisiere (String zahlAlsText)
    {
        schalteStoppuhrAn ();
        
        BigInteger zahl = new BigInteger (zahlAlsText);
        
        faktorisiere (zahl);
        
        schalteStoppuhrAus ();
    }
    
    private void faktorisiere (BigInteger zahl) 
    {
        System.out.print (zahl + " = ");
        
        BigInteger faktor;
        BigInteger rest = zahl;
        
        while (rest.compareTo (BigInteger.ONE) > 0)
        {
            faktor = gibPrimfaktor (rest);
            rest   = rest.divide (faktor);
            
            System.out.print (faktor);
            
            if (rest.compareTo (BigInteger.ONE) > 0)
            {
                System.out.print (" * ");
            }
        }
        System.out.println ();
    }
 
    /**
     * Nenne einen (naemlich den kleinsten) Primfaktor der
     * angegebenen Zahl.
     */
    public BigInteger gibPrimfaktor (String zahlAlsText)
    {
        schalteStoppuhrAn ();
        
        BigInteger zahl = new BigInteger (zahlAlsText);
        
        BigInteger antwort = gibPrimfaktor (zahl);
        
        System.out.println (zahl + " enthaelt den Primfaktor " + antwort + ".");
        schalteStoppuhrAus ();
        
        return antwort;
    }
    
    private BigInteger gibPrimfaktor (BigInteger zahl) 
    {
        /*
         * Wenn die Zahl eine bereits bekannte Primzahl ist ...
         */
        if (p.contains (zahl))
        {
            /*
             * ... dann ist sie ihr eigener Primfaktor.
             */
            return zahl;
        }
        /*
         * Ansonsten ...
         */
        else
        {
            /*
             * ... ueberpruefen wir alle bekannten Primzahlen bis zur
             * Wurzel aus der angegebenen Zahl, ob sie diese Zahl teilen.
             */
            int i = 0;
            
            while (i < p.size () && (p.get (i).pow (2)).compareTo (zahl) <= 0)
            {  
                if (zahl.remainder (p.get (i)).equals (BigInteger.ZERO))
                {
                    /*
                     * Wenn ja, geben wir den Teiler als Primfaktor zurueck.
                     */
                    return p.get (i);
                }
                
                i++;
            }
            
            /*
             * Wenn wir schon die Wurzel aus der angegebenen Zahl erreicht
             * haben ...
             */
            if (0 == i || (p.get (i-1).pow (2)).compareTo (zahl) > 0)
            {
                /*
                 * ... dann ist die Zahl prim und wir geben sie als ihren
                 * eigenen Primfaktor zurueck.
                 */
                return zahl;
            }
            /*
             * Ansonsten ...
             */
            else
            {
                /*
                 * ... muessen wir die Liste der bekannten Primzahlen erweitern.
                 * 
                 * Wir durchlaufen dadurch alle Zahlen ab der groessten bisher
                 * bekannten Primzahl bis zur Wurzel aus der angegebenen Zahl.
                 */
                for (BigInteger k = groesste.add (BigInteger.ONE); k.pow (2).compareTo (zahl) <= 0; k = k.add (BigInteger.ONE))
                {
                    /*
                     * Wenn wir dabei eine Primzahl gefunden haben ...
                     */
                    if (istPrimzahl (k))
                    {
                        /*
                         * ... merken wir sie uns und ...
                         */
                        p.add (k);
                        groesste = k;
                        
                        /*
                         * ... ueberpruefen, ob die Zahl durch die neue Primzahl teilbar ist.
                         */
                        if (zahl.remainder (k).equals (BigInteger.ZERO))
                        {
                            /*
                             * Wenn ja, ist die neue Primzahl der gesuchte Primfaktor.
                             */
                            return k;
                        }
                    }
                }
                
                /*
                 * Wenn wir bis zur Wurzel aus der angegebenen Zahl keinen Primfaktor gefunden
                 * haben, dann ist die Zahl selbst prim.
                 */
                return zahl;
            }
        }
    }
    
    /**
     * Nenne alle Primzahlen, die bisher bekannt sind.
     */
    public void nenneBekanntePrimzahlen ()
    {
        schalteStoppuhrAn ();
        
        for (int i = 0; i < p.size (); i++)
        {
            System.out.println (p.get (i));
        }
        
        schalteStoppuhrAus ();
    }
    
    /**
     * Erzeuge ein Produkt r*s aus zwei Primzahlen.
     * Der erste Faktor r soll knapp groesser als a sein.
     * Der zweite Faktor s soll knapp groesser als b sein.
     */
    public BigInteger erzeugePrimzahlprodukt (String aAlsText, String bAlsText)
    {
        schalteStoppuhrAn ();
        
        BigInteger a = new BigInteger (aAlsText);
        BigInteger b = new BigInteger (bAlsText);
        
        BigInteger r = a;
        
        while (!istPrimzahl (r))
        {
            r = r.add (BigInteger.ONE);
        }
        
        BigInteger s = b;
        
        while (!istPrimzahl (s))
        {
            s = s.add (BigInteger.ONE);
        }
        
        BigInteger produkt = r.multiply (s);
        
        System.out.println (r + " * " + s + " = " + produkt);
        
        schalteStoppuhrAus ();
        
        return produkt;
    }
    
    /**
     * Schalte die Stoppuhr an.
     */
    private void schalteStoppuhrAn ()
    {
        anfangszeit = Calendar.getInstance ();
    }
    
    /**
     * Schalte die Stoppuhr aus.
     */
    private void schalteStoppuhrAus ()
    {
        endzeit = Calendar.getInstance ();
        
        System.out.println ("Zeitbedarf: " + (endzeit.getTimeInMillis () - anfangszeit.getTimeInMillis ()) + " ms");
    }
}
